/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#define _DONT_USE_CTYPE_INLINE_	1
#include <ctype.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "pstruct.h"

/* swap a 64-bit entity */
static inline uint64_t
swap64(uint64_t u)
{
	union {
		uint64_t	u64;
		uint32_t	u32[2];
	} in, out;

	in.u64 = u;
	out.u32[0] = htonl(in.u32[1]);
	out.u32[1] = htonl(in.u32[0]);
	return out.u64;
}

/* swap a 32bit entity */
static inline uint32_t
swap32(uint32_t in)
{
	return htonl(in);
}

/* swap a 16bit entity */
static inline uint16_t
swap16(uint16_t in)
{
	return htons(in);
}

/* print a struct's fields from a file per the description */
int
pstruct(const char *description, const uint8_t *buf, size_t size, uint64_t off)
{
	const uint8_t	*bp;
	const char	*desc;
	uint64_t	 u64;
	uint32_t	 u32;
	uint16_t	 u16;
	uint32_t	 len;
	uint32_t	 i;
	char		 fmt[20];
	char		 type;

	for (desc = description, bp = &buf[off] ; *desc && *desc != '\n' && (size_t)(bp - buf) < size ; ) {
		for (len = 0 ; *desc && isdigit((uint8_t)*desc) ; desc++) {
			len = (len * 10) + (*desc - '0');
		}
		for (type = *desc++; *desc && *desc != '\n' && strchr(";:,.", *desc) == NULL ; desc++) {
			fputc(*desc, stdout);
		}
		(void) fprintf(stdout, ": ");
		if (strchr("ODXodx", type) != NULL) {
			switch(len) {
			case 8:
				(void) memcpy(&u64, bp, sizeof(u64));
				if (isupper((uint8_t)type)) {
					u64 = swap64(u64);
				}
				(void) snprintf(fmt, sizeof(fmt), "%s%%ll%c\n",
					(tolower((uint8_t)type) == 'x') ? "0x" : "",
					tolower((uint8_t)type));
				(void) fprintf(stdout, fmt, u64);
				break;
			case 4:
				(void) memcpy(&u32, bp, sizeof(u32));
				if (isupper((uint8_t)type)) {
					u32 = swap32(u32);
				}
				(void) snprintf(fmt, sizeof(fmt), "%s%%%c\n",
					(tolower((uint8_t)type) == 'x') ? "0x" : "",
					tolower((uint8_t)type));
				(void) fprintf(stdout, fmt, u32);
				break;
			case 2:
				(void) memcpy(&u16, bp, sizeof(u16));
				if (isupper((uint8_t)type)) {
					u16 = swap16(u16);
				}
				(void) snprintf(fmt, sizeof(fmt), "%s%%%c\n",
					(tolower((uint8_t)type) == 'x') ? "0x" : "",
					tolower((uint8_t)type));
				(void) fprintf(stdout, fmt, u16);
				break;
			case 1:
				(void) snprintf(fmt, sizeof(fmt), "%s%%%c\n",
					(tolower((uint8_t)type) == 'x') ? "0x" : "",
					tolower((uint8_t)type));
				(void) fprintf(stdout, fmt, (uint8_t)*bp);
				break;
			}
		} else if (type == 'h') {
			/* hexdump */
			for (i = 0 ; i < len ; i++) {
				(void) fprintf(stdout, "0x%02x%s",
					(uint8_t)*(bp + i),
					(i > 0 && i % 16 == 0) ? "\n" : " ");
			}
			(void) fprintf(stdout, "\n");
		} else {
			/* string */
			(void) fprintf(stdout, "%.*s\n", (int)len, bp);
		}
		bp += len;
		if (*desc++ == 0x0) {
			break;
		}
	}
	return 1;
}

/* print a struct's fields from a file per the description */
int
pstructf(const char *description, const char *f, uint64_t offset)
{
	uint8_t		 buf[BUFSIZ];
	FILE		*fp;
	int		 ret;
	int		 rc;

	if (f == NULL) {
		fp = stdin;
	} else if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "can't open '%s'\n", f);
		return 0;
	}
	if (offset != 0 && fseeko(fp, offset, SEEK_SET) != 0) {
		(void) fprintf(stderr, "can't seek to %" PRIu64 " '%s'\n", offset, f);
		(void) fclose(fp);
		return 0;
	}
	if ((rc = fread(buf, 1, sizeof(buf), fp)) <= 0) {
		(void) fprintf(stderr, "can't read '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	ret = pstruct(description, buf, rc, 0);
	if (f != NULL) {
		fclose(fp);
	}
	return ret;
}
